from __future__ import (
    absolute_import,
    unicode_literals,
)

from typing import (  # noqa: F401 TODO Python 3
    Any,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    TypeVar,
    Union,
)

import attr
import six


__all__ = (
    'ActionRequest',
    'ActionResponse',
    'Body',
    'Context',
    'Control',
    'JobRequest',
    'JobResponse',
    'Error',
    'UnicodeKeysDict',
)


_VT = TypeVar('_VT')
_AT = TypeVar('_AT')


class UnicodeKeysDict(dict):
    def __setitem__(self, key, value):  # type: (six.text_type, Any) -> None
        super(UnicodeKeysDict, self).__setitem__(six.text_type(key), value)

    def setdefault(self, key, default=None):  # type: (six.text_type, Optional[_VT]) -> _VT
        return super(UnicodeKeysDict, self).setdefault(six.text_type(key), default)


Body = Dict[six.text_type, Any]
Context = Dict[six.text_type, Any]
Control = Dict[six.text_type, Any]


@attr.s(frozen=True)
class Error(object):
    """The error generated by a single action."""
    code = attr.ib()  # type: six.text_type
    message = attr.ib()  # type: six.text_type
    field = attr.ib(default=None)  # type: Optional[six.text_type]
    traceback = attr.ib(default=None)  # type: Optional[six.text_type]
    variables = attr.ib(default=None)  # type: Dict[six.text_type, Any]
    denied_permissions = attr.ib(default=None)  # type: Optional[List[six.text_type]]


def _convert_errors(errors):
    # type: (Union[Iterable[Mapping[six.text_type, Any]], Iterable[Error]]) -> List[Error]
    value = []  # type: List[Error]
    for a in errors:
        value.append(a if isinstance(a, Error) else Error(**a))
    return value


@attr.s
class ActionRequest(object):
    """A request that the server execute a single action."""
    action = attr.ib()  # type: six.text_type
    body = attr.ib(default=attr.Factory(dict))  # type: Body


def _convert_action_requests(actions):
    # type: (Union[Iterable[Mapping[six.text_type, Any]], Iterable[ActionRequest]]) -> List[ActionRequest]
    value = []  # type: List[ActionRequest]
    for a in actions:
        value.append(a if isinstance(a, ActionRequest) else ActionRequest(**a))
    return value


@attr.s
class ActionResponse(object):
    """A response generated by a single action on the server."""
    action = attr.ib()  # type: six.text_type
    errors = attr.ib(default=attr.Factory(list), converter=_convert_errors)  # type: List[Error]
    body = attr.ib(default=attr.Factory(dict))  # type: Body


def _convert_action_responses(actions):
    # type: (Union[Iterable[Mapping[six.text_type, Any]], Iterable[ActionResponse]]) -> List[ActionResponse]
    value = []  # type: List[ActionResponse]
    for a in actions:
        value.append(a if isinstance(a, ActionResponse) else ActionResponse(**a))
    return value


@attr.s
class JobRequest(object):
    """
    A request that the server execute a job.

    A job consists of one or more actions and a control header. Each action is an ActionRequest,
    while the control header is a dictionary.
    """
    control = attr.ib(default=attr.Factory(dict))  # type: Control
    context = attr.ib(default=attr.Factory(dict))  # type: Context
    actions = attr.ib(default=attr.Factory(list), converter=_convert_action_requests)  # type: List[ActionRequest]


@attr.s
class JobResponse(object):
    """
    A response generated by a server job.

    Contains the result or error generated by each action in the job.
    """
    errors = attr.ib(default=attr.Factory(list), converter=_convert_errors)  # type: List[Error]
    context = attr.ib(default=attr.Factory(dict))  # type: Context
    actions = attr.ib(default=attr.Factory(list), converter=_convert_action_responses)  # type: List[ActionResponse]
